import hashlib
import re
from collections import UserDict
from functools import cached_property
from pathlib import Path

from jinja2 import Template
from passlib.handlers.sha2_crypt import sha256_crypt

from autosd_demo.cli.exceptions import ValidationError
from autosd_demo.settings import settings
from autosd_demo.utils import in_autosd_demo


class BuildEnv(UserDict):
    LINK_PATTERN = re.compile(r"^[a-z]+://")

    def __init__(self, host, osbuild=False):
        super().__init__()
        self.host = host
        self._set_build_env_config()
        if osbuild:
            self._set_osbuild_config()

    @property
    def is_remote(self):
        return True if self.host else False

    @cached_property
    def conf(self):
        return {key.lower(): val for key, val in settings.as_dict().items()}

    @cached_property
    def base_dir(self):
        if self.is_remote:
            return (
                self.conf["remote_host"][self.host]["base_dir"]
                if self.conf_exists("remote_host", self.host, "base_dir")
                else "~"
            )

        return self.conf["base_dir"]

    @property
    def build_id(self):
        return self["build"]["build_id"]

    @property
    def hooks(self):
        return self["hooks"]

    @cached_property
    def build_dir(self):
        if self.is_remote:
            return (
                self.conf["remote_host"][self.host]["build_dir"]
                if self.conf_exists("remote_host", self.host, "build_dir")
                else str(Path(self.base_dir) / "build")
            )

        return self.conf["build_dir"]

    def conf_exists(self, *keys):
        return self.key_exists(self.conf, *keys)

    @staticmethod
    def key_exists(element, *keys):
        _element = element
        for key in keys:
            try:
                _element = _element[key]
            except KeyError:
                return False
        return True

    def conf_required(self, section: str, key: str, *option):
        if not self.conf_exists(section, key, *option):
            option_name = ".".join(option)
            raise ValidationError(section, key, f"'{option_name}' is missing")

    def __repr__(self):
        return f"{type(self).__name__}({self.data})"

    def _set_build_env_config(self):
        build_env = {"conf": self.conf, "build": {}, "hooks": {}}
        build_env["build"]["is_remote"] = self.is_remote
        if in_autosd_demo():
            build_env["build"]["base_dir"] = self.base_dir
            build_env["build"]["build_dir"] = self.build_dir
            build_env["build"]["src"] = self._get_build_src_contents()
            custom_containers_ids = self._get_build_custom_containers_ids()
            build_env["build"]["custom_containers"] = custom_containers_ids
            build_env["build"]["podman_extra_args"] = self._get_podman_extra_args()
            build_env["hooks"] = self._get_configured_hooks()
        self.update(build_env)

    def _get_configured_hooks(self):
        hooks = {"pre_build": [], "post_build": [], "clean_build": [], "setup": []}
        if self.conf_exists("hooks"):
            for hook_data in self.conf["hooks"].values():
                stages = ["setup", "pre_build", "post_build", "clean_build"]
                for stage in stages:
                    if stage in hook_data:
                        for hook in hook_data[stage]:
                            hook_path = (
                                hook
                                if Path(hook).is_absolute()
                                else str(Path(self.base_dir) / hook)
                            )
                            hooks[stage].append(hook_path)

        for stage in hooks.keys():
            hooks[stage] = sorted(hooks[stage], key=lambda path: Path(path).name)

        return hooks

    def reload_src_contents(self):
        self["build"]["src"] = self._get_build_src_contents()

    def _get_build_src_contents(self):
        src_contents = {
            "directories": set(),
            "files": [],
            "templated_files": [],
            "fetch_files": [],
        }

        self._get_build_src_quadlets(src_contents)
        self._get_build_src_extra_files(src_contents)
        self._get_build_src_custom_containers_files(src_contents)

        src_contents["directories"] = reduce_directory_list(src_contents["directories"])

        return src_contents

    def _get_build_custom_containers_ids(self):
        result = []
        if self.conf_exists("containers"):
            for container_id, container_data in self.conf["containers"].items():
                if (
                    "container_context" in container_data
                    or "container_file" in container_data
                ):
                    result.append(container_id)
        return result

    def _get_build_src_custom_containers_files(self, src_contents):
        if self.conf_exists("containers"):
            for container_id, container_data in self.conf["containers"].items():
                if "container_context" in container_data:
                    context_path = Path(container_data["container_context"])
                    if not context_path.is_absolute():
                        context_path = Path(self.conf["base_dir"]) / context_path

                    list_files = [f for f in context_path.glob("**/*") if f.is_file()]

                    for file in list_files:
                        path_template = Template(
                            file.as_posix(),
                            variable_start_string="{",
                            variable_end_string="}",
                        )
                        src = Path(path_template.render(conf=self.conf))
                        dest = Path("files") / self.conf["name"]
                        dest = dest / "containers" / container_id
                        dest = dest / src.relative_to(context_path)

                        if file.suffix == ".j2":
                            src_contents["templated_files"].append(
                                {
                                    "src": file.as_posix(),
                                    "dest": dest.with_suffix("").as_posix(),
                                }
                            )
                        else:
                            src_contents["files"].append(
                                {"src": file.as_posix(), "dest": dest.as_posix()}
                            )
                        src_contents["directories"].add(dest.parent.as_posix())

        return src_contents

    def _get_build_src_extra_files(self, src_contents):
        if self.conf_exists("extra_files"):
            for ef_id, ef_data in self.conf["extra_files"].items():
                touch = ef_data.get("touch", False)
                if not touch:
                    self.conf_required("extra_files", ef_id, "src")
                self.conf_required("extra_files", ef_id, "dest")

                dest_path = (
                    Path(ef_data["dest"][1:])
                    if Path(ef_data["dest"]).is_absolute()
                    else Path(ef_data["dest"])
                )
                if touch:
                    self._get_build_src_extra_files_touch(
                        ef_data, dest_path, src_contents
                    )
                elif self.LINK_PATTERN.match(ef_data["src"]):
                    self._get_build_src_extra_files_link(
                        ef_data, dest_path, src_contents
                    )
                else:
                    self._get_build_src_extra_files_regular(
                        ef_data, dest_path, src_contents
                    )

    def _get_build_src_extra_files_touch(self, ef_data, dest_path, src_contents):
        layer = ef_data.get("layer", "root")

        dest = Path("files") / self.conf["name"]
        dest = dest / "extra_files" / f"{layer}_fs"
        dest = dest / dest_path

        if ef_data["dest"].endswith("/"):
            src_contents["directories"].add(dest.as_posix())
        else:
            src_contents["files"].append({"dest": dest.as_posix()})
            src_contents["directories"].add(dest.parent.as_posix())

    def _get_build_src_extra_files_link(self, ef_data, dest_path, src_contents):
        layer = ef_data.get("layer", "root")

        url = ef_data["src"]
        checksum = ef_data["checksum"] if "checksum" in ef_data else ""
        file_name = Path(url.replace("://", "_").replace("/", "_"))
        src = Path(self.conf["base_dir"]) / "cache" / "downloads" / file_name
        dest = (
            Path("files")
            / self.conf["name"]
            / "extra_files"
            / f"{layer}_fs"
            / dest_path
        )
        if ef_data["dest"].endswith("/"):
            dest_name = url.rpartition("/")[2]
            dest = dest / dest_name

        src_contents["fetch_files"].append(
            {"url": url, "checksum": checksum, "file": file_name.as_posix()}
        )
        src_contents["files"].append({"src": src.as_posix(), "dest": dest.as_posix()})
        src_contents["directories"].add(dest.parent.as_posix())

    def _get_build_src_extra_files_regular(self, ef_data, dest_path, src_contents):
        layer = ef_data.get("layer", "root")
        src_path = Path(ef_data["src"])

        if not src_path.is_absolute():
            src_path = Path(self.conf["base_dir"]) / src_path

        if src_path.is_file():
            list_files = [src_path]
        elif src_path.is_dir():
            list_files = [f for f in src_path.glob("**/*") if f.is_file()]
        else:
            return

        for file in list_files:
            path_template = Template(
                file.as_posix(),
                variable_start_string="{",
                variable_end_string="}",
            )
            src = Path(path_template.render(conf=self.conf))
            dest = Path("files") / self.conf["name"]
            dest = dest / "extra_files" / f"{layer}_fs"
            dest = dest / dest_path

            if len(list_files) == 1 and ef_data["dest"].endswith("/"):
                dest = dest / src.name
            else:
                dest = dest / src.relative_to(src_path)

            if file.suffix == ".j2":
                if (
                    len(list_files) > 1
                    or not src_path.is_file()
                    or ef_data["dest"].endswith("/")
                ):
                    dest = dest.with_suffix("")
                src_contents["templated_files"].append(
                    {
                        "src": file.as_posix(),
                        "dest": dest.as_posix(),
                    }
                )
            else:
                src_contents["files"].append(
                    {"src": file.as_posix(), "dest": dest.as_posix()}
                )
            src_contents["directories"].add(dest.parent.as_posix())

    def _get_build_src_quadlets(self, src_contents):
        if self.conf_exists("containers"):
            for container_id, container_data in self.conf["containers"].items():
                if not container_data.get("quadlet", True):
                    continue

                layer = (
                    "root" if "layer" not in container_data else container_data["layer"]
                )
                quadlet_dest = (
                    Path("files") / self.conf["name"] / "extra_files" / f"{layer}_fs"
                )
                quadlet_dest = (
                    quadlet_dest / f"etc/containers/systemd/{container_id}.container"
                )

                quadlet_template = {
                    "src": self.conf["centos_sig"]["autosd_demo"][
                        "template_systemd_container_path"
                    ],
                    "dest": quadlet_dest.as_posix(),
                    "container_name": container_id,
                }

                src_contents["templated_files"].append(quadlet_template)
                src_contents["directories"].add(quadlet_dest.parent.as_posix())

    def _get_podman_extra_args(self):
        extra_args = {}
        if self.conf_exists("containers"):
            for container_id, container_data in self.conf["containers"].items():
                if "build_args" in container_data:
                    extra_args[container_id] = [
                        f"--build-arg {key}={value}"
                        for key, value in container_data["build_args"].items()
                    ]
        return extra_args

    def _set_osbuild_config(self):
        osbuild_conf = {}
        osbuild_conf.update(self._get_osbuild_preprocessed_containers())
        osbuild_conf.update(self._get_osbuild_preprocessed_users())
        osbuild_conf.update(self._get_osbuild_preprocessed_groups())
        osbuild_conf.update(self._get_osbuild_preprocessed_permissions())
        osbuild_conf.update(self._get_osbuild_preprocessed_systemd())
        osbuild_conf.update(self._get_osbuild_preprocessed_extra_contents())
        osbuild_conf["has_qm"] = any(
            v for k, v in osbuild_conf.items() if k.startswith("qm_")
        )
        # mpp_vars are processed last
        osbuild_conf.update(self._get_osbuild_preprocessed_mpp_vars(osbuild_conf))
        self.update({"osbuild": osbuild_conf})

    def _get_osbuild_preprocessed_systemd(self):
        osbuild_conf = {"systemd": {}, "qm_systemd": {}}
        special_params = ["layer"]
        allowed_params = ["enabled_services", "disabled_services"]
        if self.conf_exists("systemd"):
            for systemd_id, systemd_data in self.conf["systemd"].items():
                systemd_v = (
                    osbuild_conf["qm_systemd"]
                    if "layer" in systemd_data and systemd_data["layer"] == "qm"
                    else osbuild_conf["systemd"]
                )
                for key, value in systemd_data.items():
                    if key in special_params:
                        continue
                    if key not in allowed_params:
                        raise ValidationError(
                            "systemd", systemd_id, f"option '{key}' is not allowed"
                        )

                    systemd_v.setdefault(key, [])
                    if isinstance(value, list):
                        systemd_v[key] = list(set(systemd_v[key] + value))
                    elif isinstance(value, str):
                        if value not in systemd_v[key]:
                            systemd_v[key].append(value)
                    else:
                        raise ValidationError(
                            "systemd",
                            systemd_id,
                            f"'{key}' value must be string or list",
                        )
                    systemd_v[key].sort()

        return osbuild_conf

    def _get_osbuild_preprocessed_users(self):
        osbuild_conf = {"users": {}, "qm_users": {}}
        special_params = ["layer", "name"]
        allowed_params = ["uid", "password", "gid", "groups", "home", "description"]
        if self.conf_exists("users"):
            for user_id, user_data in self.conf["users"].items():
                username = user_data["name"] if "name" in user_data else user_id
                new_user = {f"{username}": {}}
                for key, value in user_data.items():
                    if key in special_params:
                        continue
                    if key not in allowed_params:
                        raise ValidationError(
                            "users", user_id, f"option '{key}' is not allowed"
                        )

                    if key == "password":
                        new_user[username][key] = sha256_crypt.hash(value)
                    elif key in ["uid", "gid"]:
                        new_user[username][key] = int(value)
                    elif key == "groups" and not isinstance(value, list):
                        new_user[username][key] = [value]
                    else:
                        new_user[username][key] = value

                users_v = (
                    osbuild_conf["qm_users"]
                    if "layer" in user_data and user_data["layer"] == "qm"
                    else osbuild_conf["users"]
                )
                users_v.update(new_user)

        return osbuild_conf

    def _get_osbuild_preprocessed_groups(self):
        osbuild_conf = {"groups": {}, "qm_groups": {}}
        special_params = ["layer"]
        allowed_params = ["gid"]
        if self.conf_exists("groups"):
            for group_id, group_data in self.conf["groups"].items():
                groupname = group_data["name"] if "name" in group_data else group_id
                new_group = {f"{groupname}": {}}
                for key, value in group_data.items():
                    if key in special_params:
                        continue
                    if key not in allowed_params:
                        raise ValidationError(
                            "groups", group_id, f"option '{key}' is not allowed"
                        )

                    if key == "gid":
                        new_group[groupname][key] = int(value)

                groups_v = (
                    osbuild_conf["qm_groups"]
                    if "layer" in group_data and group_data["layer"] == "qm"
                    else osbuild_conf["groups"]
                )
                groups_v.update(new_group)

        return osbuild_conf

    def _get_osbuild_preprocessed_permissions(self):
        osbuild_conf = {"chown": {}, "chmod": {}, "qm_chown": {}, "qm_chmod": {}}

        if self.conf_exists("fs_perm"):
            for fs_perm_id, fs_perm_data in self.conf["fs_perm"].items():
                self.conf_required("fs_perm", fs_perm_id, "path")

                (chown_v, chmod_v) = (
                    (osbuild_conf["qm_chown"], osbuild_conf["qm_chmod"])
                    if "layer" in fs_perm_data and fs_perm_data["layer"] == "qm"
                    else (osbuild_conf["chown"], osbuild_conf["chmod"])
                )
                if "user" in fs_perm_data or "group" in fs_perm_data:
                    chown = {}
                    if "user" in fs_perm_data:
                        chown["user"] = fs_perm_data["user"]
                    if "group" in fs_perm_data:
                        chown["group"] = fs_perm_data["group"]
                    chown["recursive"] = fs_perm_data.get("recursive", False)
                    chown_v[fs_perm_data["path"]] = chown
                if "perms" in fs_perm_data:
                    if not isinstance(fs_perm_data["perms"], str):
                        raise ValidationError(
                            "fs_perm", fs_perm_id, "'perms' value must be string"
                        )
                    chmod = {
                        "mode": fs_perm_data["perms"],
                        "recursive": fs_perm_data.get("recursive", False),
                    }
                    chmod_v[fs_perm_data["path"]] = chmod

        return osbuild_conf

    def _get_osbuild_preprocessed_containers(self):
        osbuild_conf = {"containers": [], "qm_containers": []}

        if self.conf_exists("containers"):
            for container_id, container_data in self.conf["containers"].items():
                self.conf_required(
                    "containers", container_id, "systemd", "container", "image"
                )
                new_container = {
                    "tag": container_data.get("version", "latest"),
                    "name": container_data["systemd"]["container"]["image"],
                }
                image_name = container_data.get("image", container_id)

                if set(container_data.keys()).intersection(
                    ["container_context", "container_file"]
                ):
                    new_container["source"] = f"localhost/{image_name}"
                    new_container["containers-transport"] = "containers-storage"
                else:
                    source = [
                        container_data.get("registry", None),
                        container_data.get("namespace", None),
                        image_name,
                    ]
                    source = [x for x in source if x is not None]
                    new_container["source"] = "/".join(source)

                if "layer" in container_data and container_data["layer"] == "qm":
                    osbuild_conf["qm_containers"].append(new_container)
                else:
                    osbuild_conf["containers"].append(new_container)

        return osbuild_conf

    def _get_osbuild_preprocessed_extra_contents(self):
        osbuild_conf = {"extra_contents": {}, "qm_extra_contents": {}}

        self._osbuild_preprocess_extra_contents_files(osbuild_conf)
        self._osbuild_preprocess_extra_contents_directories(osbuild_conf)
        self._osbuild_finalize_extra_contents(osbuild_conf)

        return osbuild_conf

    def _osbuild_preprocess_extra_contents_files(self, osbuild_conf):
        extra_contents_list = self["build"]["src"]["files"]
        extra_contents_list += self["build"]["src"]["templated_files"]
        extra_contents_list = sorted(extra_contents_list, key=lambda x: x["dest"])

        for item in extra_contents_list:
            dest_path = Path(item["dest"])
            if dest_path.parts[2] != "extra_files":
                continue
            layer = dest_path.parts[3].split("_")[0]

            extra_contents_v = (
                osbuild_conf["qm_extra_contents"]
                if layer == "qm"
                else osbuild_conf["extra_contents"]
            )

            item_idx = extra_contents_v.setdefault("index", 0)
            content_id = hashlib.sha256(
                f"{item.get('src', '')}\n{item['dest']}".encode()
            ).hexdigest()[:7]
            tree_dest_path = dest_path.relative_to(Path(*dest_path.parts[:4]))

            ec_input = {
                f"{layer}_extra_content_{item_idx}": {
                    "type": "org.osbuild.files",
                    "origin": "org.osbuild.source",
                    "mpp-embed": {
                        "id": f"{content_id}",
                        "path": Path("..").joinpath(dest_path).as_posix(),
                    },
                }
            }

            ec_path = {
                "from": {
                    "mpp-format-string": (
                        f"input://{layer}_extra_content_{item_idx}/{{embedded['{content_id}']}}"
                    )
                },
                "to": f"tree:///{tree_dest_path.as_posix()}",
            }

            extra_contents_v.setdefault("inputs", {}).update(ec_input)
            extra_contents_v.setdefault("paths", []).append(ec_path)
            extra_contents_v.setdefault("directories", []).append(
                "/" + tree_dest_path.parent.as_posix()
            )
            extra_contents_v["index"] += 1

    def _osbuild_preprocess_extra_contents_directories(self, osbuild_conf):
        for directory in self["build"]["src"].get("directories", []):
            dir_path = Path(directory)
            if dir_path.parts[2] != "extra_files":
                continue
            layer = dir_path.parts[3].split("_")[0]

            extra_contents_v = (
                osbuild_conf["qm_extra_contents"]
                if layer == "qm"
                else osbuild_conf["extra_contents"]
            )

            tree_dest_path = dir_path.relative_to(Path(*dir_path.parts[:4]))
            extra_contents_v.setdefault("inputs", {})
            extra_contents_v.setdefault("paths", [])
            extra_contents_v.setdefault("directories", []).append(
                "/" + tree_dest_path.as_posix()
            )

    def _osbuild_finalize_extra_contents(self, osbuild_conf):
        for extra_contents_v in [
            osbuild_conf["extra_contents"],
            osbuild_conf["qm_extra_contents"],
        ]:
            if "index" in extra_contents_v:
                extra_contents_v.pop("index")

            if "directories" in extra_contents_v:
                extra_contents_v["directories"] = [
                    {"path": d, "exist_ok": True, "parents": True}
                    for d in reduce_directory_list(extra_contents_v["directories"])
                ]

    def _get_osbuild_preprocessed_mpp_vars(self, conf):
        osbuild_conf = {"mpp_vars": self.conf.get("mpp_vars", {})}
        if self.conf_exists("rpm"):
            osbuild_conf["mpp_vars"].update(self._get_osbuild_preprocessed_rpms())
        if self.conf_exists("rpm_repo"):
            osbuild_conf["mpp_vars"].update(self._get_osbuild_preprocessed_rpm_repos())
        if self.conf_exists("dracut"):
            osbuild_conf["mpp_vars"].update(
                self._get_osbuild_preprocessed_dracut_conf()
            )
        if conf.get("containers", False):
            osbuild_conf["mpp_vars"].update({"use_containers_extra_store": True})
        if conf.get("qm_containers", False):
            osbuild_conf["mpp_vars"].update({"use_qm_containers_extra_store": True})
        if not conf.get("has_qm", False):
            conf["has_qm"] = any(k.startswith("qm_") for k in osbuild_conf["mpp_vars"])
        if conf["has_qm"]:
            osbuild_conf["mpp_vars"].update({"use_qm": True})
            osbuild_conf["mpp_vars"].update({"use_bluechi": True})
        return osbuild_conf

    def _get_osbuild_preprocessed_dracut_conf(self):
        osbuild_conf = {}
        for dracut_data in self.conf["dracut"].values():
            if "modules" in dracut_data:
                osbuild_conf.setdefault("dracut_add_modules", []).extend(
                    dracut_data["modules"]
                )
            if "install" in dracut_data:
                osbuild_conf.setdefault("dracut_install", []).extend(
                    dracut_data["install"]
                )
        return osbuild_conf

    def _get_osbuild_preprocessed_rpm_repos(self):
        osbuild_conf = {}
        for rpm_repo_id, rpm_repo_data in self.conf["rpm_repo"].items():
            self.conf_required("rpm_repo", rpm_repo_id, "baseurl")

            new_repo = {"id": rpm_repo_id, "baseurl": rpm_repo_data["baseurl"]}

            layer_key = "extra_repos"
            if rpm_repo_data.get("layer") == "qm":
                layer_key = f"qm_{layer_key}"
            osbuild_conf.setdefault(layer_key, []).append(new_repo)

        return osbuild_conf

    def _get_osbuild_preprocessed_rpms(self):
        osbuild_conf = {}
        for rpm_id, rpm_data in self.conf["rpm"].items():
            self.conf_required("rpm", rpm_id, "packages")

            layer_key = "extra_rpms"
            if rpm_data.get("layer") == "qm":
                layer_key = f"qm_{layer_key}"
            osbuild_conf.setdefault(layer_key, []).extend(rpm_data["packages"])

        return osbuild_conf


def reduce_directory_list(directories):
    directory_set = directories if isinstance(directories, set) else set(directories)
    sorted_list = sorted(list(directory_set))
    dir_paths = [Path(d) for d in sorted_list]
    reduced_directory_list = []
    for index, dir_x in enumerate(dir_paths):
        if all(dir_x not in dir_y.parents for dir_y in dir_paths[index + 1 :]):
            reduced_directory_list.append(dir_x.as_posix())

    return reduced_directory_list
